"""
@Author = 'Michael Stanley'

Tests for RivendellAudioImporter.py

============ Change Log ============
2023-Jan-20 = Change license from GPLv2 to GPLv3.

2023-Jan-19 = Fix an improperly failing test.

2022-May-09 = Backported to Python 3.6, since that is what is available on 
                CentOS 7.

2020-Jul-25 = Added GPL notice.

2020-Jul-01 = Created.

============ License ============
Copyright (C) 2020, 2022-2023 Michael Stanley

This file is part of wmul_rivendell.

wmul_rivendell is free software: you can redistribute it and/or modify it 
under the terms of the GNU General Public License as published by the Free 
Software Foundation, either version 3 of the License, or (at your option) any 
later version.

wmul_rivendell is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with 
wmul_rivendell. If not, see <https://www.gnu.org/licenses/>. 
"""
import pytest
import wmul_test_utils

from collections import namedtuple
from wmul_rivendell import RivendellAudioImporter

run_script_options = namedtuple(
    "run_script_options",
    [
        "use_log_syslog",
        "use_log_filename"
    ]
)
run_script_params, run_script_ids = \
    wmul_test_utils.generate_true_false_matrix_from_namedtuple(run_script_options)


@pytest.fixture(scope="function", params=run_script_params, ids=run_script_ids)
def setup_run_script(request, mocker, caplog):
    params = request.param
    mock_cache_duration = 100
    mock_source_paths = ["mock_source_paths"]

    if params.use_log_syslog:
        mock_rdimport_syslog = True
        mock_rdimport_log_file_name = None
        expected_log_argument = "--log-syslog"
    elif params.use_log_filename:
        mock_rdimport_syslog = False
        mock_rdimport_log_file_name = "mock_rdimport_log_file_name"
        expected_log_argument = f"--log-filename={mock_rdimport_log_file_name}"
    else:
        mock_rdimport_syslog = False
        mock_rdimport_log_file_name = None
        expected_log_argument = None

    mock_arguments = RivendellAudioImporter.ImportRivendellFileWithFileSystemMetadataArguments(
        source_paths=mock_source_paths,
        cache_duration=mock_cache_duration,
        rdimport_syslog=mock_rdimport_syslog,
        rdimport_log_file_name=mock_rdimport_log_file_name
    )

    mock_ttl_cache = "mock_ttl_cache"
    mock_ttl_cache_constructor = mocker.Mock(return_value=mock_ttl_cache)
    mocker.patch("wmul_rivendell.RivendellAudioImporter.cachetools.TTLCache",
                  mock_ttl_cache_constructor)

    mock_returned_files_for_importer = "mock_returned_files_for_importer"

    mock_gather_file_names = mocker.Mock(return_value=(None, mock_returned_files_for_importer))
    mocker.patch("wmul_rivendell.RivendellAudioImporter._gather_file_names",
                 mock_gather_file_names)

    mock_run_importer = mocker.Mock()
    mocker.patch("wmul_rivendell.RivendellAudioImporter._run_importer_on_selected_files",
                 mock_run_importer)

    def mock_sleep_function(dur):
        raise RuntimeError
    mock_sleep = mocker.Mock(side_effect=mock_sleep_function)

    mocker.patch("wmul_rivendell.RivendellAudioImporter.sleep",
                 mock_sleep)

    expected_log_messages = [
        f"With {mock_arguments}",
        "Loop!"
    ]

    with pytest.raises(RuntimeError):
        RivendellAudioImporter.run_script(mock_arguments)

    return wmul_test_utils.make_namedtuple(
        "setup_run_script",
        mock_arguments=mock_arguments,
        mock_ttl_cache_constructor=mock_ttl_cache_constructor,
        mock_cache_duration=mock_cache_duration,
        expected_log_messages=expected_log_messages,
        caplog_text=caplog.text,
        mock_gather_file_names=mock_gather_file_names,
        mock_source_paths=mock_source_paths,
        mock_ttl_cache=mock_ttl_cache,
        mock_run_importer=mock_run_importer,
        mock_returned_files_for_importer=mock_returned_files_for_importer,
        expected_log_argument=expected_log_argument,
        mock_sleep=mock_sleep
    )


def test_run_script(setup_run_script, mocker):
    mock_ttl_cache_constructor_expected_calls = [
        mocker.call(10_000, setup_run_script.mock_cache_duration),
        mocker.call(5, 3600)
    ]

    wmul_test_utils.assert_has_only_these_calls(
        setup_run_script.mock_ttl_cache_constructor, 
        mock_ttl_cache_constructor_expected_calls
    )

    for elm in setup_run_script.expected_log_messages:
        assert elm in setup_run_script.caplog_text

    setup_run_script.mock_gather_file_names.assert_called_once()

    mgfn_call_source_paths, mgfn_call_previously_seen_files, mgfn_call_previously_sent_to_importer_cache = \
        setup_run_script.mock_gather_file_names.call_args[0]

    assert mgfn_call_source_paths == setup_run_script.mock_source_paths
    assert isinstance(mgfn_call_previously_seen_files, dict)
    assert mgfn_call_previously_sent_to_importer_cache == setup_run_script.mock_ttl_cache

    setup_run_script.mock_run_importer.assert_called_once_with(
        setup_run_script.mock_returned_files_for_importer,
        setup_run_script.mock_ttl_cache,
        setup_run_script.expected_log_argument
    )

    setup_run_script.mock_sleep.assert_called_once_with(10)



"""
@Author = 'Michael Stanley'

============ Change Log ============
2023-Jan-20 = Change license from GPLv2 to GPLv3.

2021-Mar-11 = Created.

============ License ============
Copyright (C) 2021, 2023 Michael Stanley

This file is part of wmul_rivendell.

wmul_rivendell is free software: you can redistribute it and/or modify it 
under the terms of the GNU General Public License as published by the Free 
Software Foundation, either version 3 of the License, or (at your option) any 
later version.

wmul_rivendell is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with 
wmul_rivendell. If not, see <https://www.gnu.org/licenses/>. 
"""
from wmul_rivendell.LoadCurrentLogLine import _get_start_time_in_millis
from datetime import datetime


def test_zero_time():
    time_under_test = datetime(year=2020, month=1, day=1, hour=0, minute=0, second=0)

    expected_time_in_millis = 0
    result_time_in_millis = _get_start_time_in_millis(time_under_test)

    assert expected_time_in_millis == result_time_in_millis


def test_3_50_20_am():
    time_under_test = datetime(year=2020, month=1, day=1, hour=3, minute=50, second=20)

    expected_time_in_millis = 13_820_000
    result_time_in_millis = _get_start_time_in_millis(time_under_test)

    assert expected_time_in_millis == result_time_in_millis


def test_3_50_20_pm():
    time_under_test = datetime(year=2020, month=1, day=1, hour=15, minute=50, second=20)

    expected_time_in_millis = 57_020_000
    result_time_in_millis = _get_start_time_in_millis(time_under_test)

    assert expected_time_in_millis == result_time_in_millis


def test_two_days_same_hour_same_millis():
    first_time_under_test = datetime(year=2020, month=1, day=1, hour=15, minute=50, second=0)
    second_time_under_test = datetime(year=2020, month=1, day=2, hour=15, minute=50, second=0)

    first_time_in_millis = _get_start_time_in_millis(first_time_under_test)
    second_time_in_millis = _get_start_time_in_millis(second_time_under_test)

    assert first_time_in_millis == second_time_in_millis



"""
@Author = 'Michael Stanley'

============ Change Log ============
2023-May-25 = Moved from tests/FilterCartReportForMusicScheduler when the 
                code under test was refactored into LoadCartDataDump.

2023-Jan-20 = Change license from GPLv2 to GPLv3.

2020-Oct-21 = Created.

============ License ============
Copyright (C) 2020, 2023, 2025 Michael Stanley

This file is part of wmul_rivendell.

wmul_rivendell is free software: you can redistribute it and/or modify it 
under the terms of the GNU General Public License as published by the Free 
Software Foundation, either version 3 of the License, or (at your option) any 
later version.

wmul_rivendell is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with 
wmul_rivendell. If not, see <https://www.gnu.org/licenses/>. 
"""
from collections import namedtuple
from wmul_rivendell.LoadCartDataDump import CartType, RivendellCart
import pytest
import wmul_test_utils

from_dict_options = namedtuple(
    "from_dict_options",
    [
        "audio_cart"
    ]
)


from_dict_params, from_dict_ids = wmul_test_utils.generate_true_false_matrix_from_namedtuple(from_dict_options)


@pytest.fixture(scope="function", params=from_dict_params, ids=from_dict_ids)
def setup_from_dict(request):
    if request.param.audio_cart:
        expected_type = CartType.Audio
        type_for_dict = "audio"
    else:
        expected_type = CartType.Macro
        type_for_dict = "macro"

    expected_cart_number = '638525'
    expected_cut_number = '1'
    expected_group_name = '"EXPLICABO"'
    expected_title = 'perferendis enim harum repudiandae'
    expected_artist = 'nisi ipsum'
    expected_album = 'molestiae aut'
    expected_year = '"2017"'
    expected_isrc = 'atque ratione'
    expected_isci = 'consequuntur eligendi'
    expected_label = 'occaecati vero'
    expected_client = 'amet fugiat'
    expected_agency = 'possimus corrupti'
    expected_publisher = 'soluta rerum'
    expected_composer = '"placeat sed'
    expected_conductor = 'enim architecto'
    expected_song_id = 'suscipit veritatis'
    expected_user_defined = 'accusamus reprehenderit'
    expected_description = 'repellat consectetur'
    expected_outcue = 'maiores incidunt'
    expected_filename = 'recusandae unde'
    expected_length = '7:14'
    expected_length_in_seconds = 434
    expected_start_point = 'repudiandae totam'
    expected_end_point = 'aut iste'
    expected_segue_start_point = 'dignissimos tenetur'
    expected_segue_end_point = 'molestiae doloremque'
    expected_hook_start_point = 'doloribus cupiditate'
    expected_hook_end_point = 'reprehenderit perspiciatis'
    expected_talk_start_point = 'similique dolorum'
    expected_talk_end_point = 'dolorem nostrum'
    expected_fadeup_point = 'assumenda praesentium'
    expected_fadedown_point = 'minus vitae'
    expected_sched_codes = 'non officia'

    source_dict = {
        "CART_NUMBER": expected_cart_number,
        "CUT_NUMBER": expected_cut_number,
        "TYPE": type_for_dict,
        "GROUP_NAME": expected_group_name,
        "TITLE": expected_title,
        "ARTIST": expected_artist,
        "ALBUM": expected_album,
        "YEAR": expected_year,
        "ISRC": expected_isrc,
        "ISCI": expected_isci,
        "LABEL": expected_label,
        "CLIENT": expected_client,
        "AGENCY": expected_agency,
        "PUBLISHER": expected_publisher,
        "COMPOSER": expected_composer,
        "CONDUCTOR": expected_conductor,
        "SONG_ID": expected_song_id,
        "USER_DEFINED": expected_user_defined,
        "DESCRIPTION": expected_description,
        "OUTCUE": expected_outcue,
        "FILENAME": expected_filename,
        "LENGTH": expected_length,
        "START_POINT": expected_start_point,
        "END_POINT": expected_end_point,
        "SEGUE_START_POINT": expected_segue_start_point,
        "SEGUE_END_POINT": expected_segue_end_point,
        "HOOK_START_POINT": expected_hook_start_point,
        "HOOK_END_POINT": expected_hook_end_point,
        "TALK_START_POINT": expected_talk_start_point,
        "TALK_END_POINT": expected_talk_end_point,
        "FADEUP_POINT": expected_fadeup_point,
        "FADEDOWN_POINT": expected_fadedown_point,
        "SCHED_CODES": expected_sched_codes
    }

    result_rivendell_cart = RivendellCart.from_dict(source_dict=source_dict)

    return wmul_test_utils.make_namedtuple(
        "setup_from_dict__audio",
        expected_cart_number=expected_cart_number,
        expected_cut_number=expected_cut_number,
        expected_type=expected_type,
        expected_group_name=expected_group_name,
        expected_title=expected_title,
        expected_artist=expected_artist,
        expected_album=expected_album,
        expected_year=expected_year,
        expected_isrc=expected_isrc,
        expected_isci=expected_isci,
        expected_label=expected_label,
        expected_client=expected_client,
        expected_agency=expected_agency,
        expected_publisher=expected_publisher,
        expected_composer=expected_composer,
        expected_conductor=expected_conductor,
        expected_song_id=expected_song_id,
        expected_user_defined=expected_user_defined,
        expected_description=expected_description,
        expected_outcue=expected_outcue,
        expected_filename=expected_filename,
        expected_length=expected_length,
        expected_length_in_seconds=expected_length_in_seconds,
        expected_start_point=expected_start_point,
        expected_end_point=expected_end_point,
        expected_segue_start_point=expected_segue_start_point,
        expected_segue_end_point=expected_segue_end_point,
        expected_hook_start_point=expected_hook_start_point,
        expected_hook_end_point=expected_hook_end_point,
        expected_talk_start_point=expected_talk_start_point,
        expected_talk_end_point=expected_talk_end_point,
        expected_fadeup_point=expected_fadeup_point,
        expected_fadedown_point=expected_fadedown_point,
        expected_sched_codes=expected_sched_codes,
        result_rivendell_cart=result_rivendell_cart
    )


def test_from_dict_cart_number_correct(setup_from_dict):
    result_rivendell_cart = setup_from_dict.result_rivendell_cart
    assert setup_from_dict.expected_cart_number == result_rivendell_cart.cart_number


def test_from_dict_cut_number_correct(setup_from_dict):
    result_rivendell_cart = setup_from_dict.result_rivendell_cart
    assert setup_from_dict.expected_cut_number == result_rivendell_cart.cut_number


def test_from_dict_type_correct(setup_from_dict):
    result_rivendell_cart = setup_from_dict.result_rivendell_cart
    assert setup_from_dict.expected_type == result_rivendell_cart.type


def test_from_dict_group_name_correct(setup_from_dict):
    result_rivendell_cart = setup_from_dict.result_rivendell_cart
    assert setup_from_dict.expected_group_name == result_rivendell_cart.group_name


def test_from_dict_title_correct(setup_from_dict):
    result_rivendell_cart = setup_from_dict.result_rivendell_cart
    assert setup_from_dict.expected_title == result_rivendell_cart.title


def test_from_dict_artist_correct(setup_from_dict):
    result_rivendell_cart = setup_from_dict.result_rivendell_cart
    assert setup_from_dict.expected_artist == result_rivendell_cart.artist


def test_from_dict_album_correct(setup_from_dict):
    result_rivendell_cart = setup_from_dict.result_rivendell_cart
    assert setup_from_dict.expected_album == result_rivendell_cart.album


def test_from_dict_year_correct(setup_from_dict):
    result_rivendell_cart = setup_from_dict.result_rivendell_cart
    assert setup_from_dict.expected_year == result_rivendell_cart.year


def test_from_dict_isrc_correct(setup_from_dict):
    result_rivendell_cart = setup_from_dict.result_rivendell_cart
    assert setup_from_dict.expected_isrc == result_rivendell_cart.isrc


def test_from_dict_isci_correct(setup_from_dict):
    result_rivendell_cart = setup_from_dict.result_rivendell_cart
    assert setup_from_dict.expected_isci == result_rivendell_cart.isci


def test_from_dict_label_correct(setup_from_dict):
    result_rivendell_cart = setup_from_dict.result_rivendell_cart
    assert setup_from_dict.expected_label == result_rivendell_cart.label


def test_from_dict_client_correct(setup_from_dict):
    result_rivendell_cart = setup_from_dict.result_rivendell_cart
    assert setup_from_dict.expected_client == result_rivendell_cart.client


def test_from_dict_agency_correct(setup_from_dict):
    result_rivendell_cart = setup_from_dict.result_rivendell_cart
    assert setup_from_dict.expected_agency == result_rivendell_cart.agency


def test_from_dict_publisher_correct(setup_from_dict):
    result_rivendell_cart = setup_from_dict.result_rivendell_cart
    assert setup_from_dict.expected_publisher == result_rivendell_cart.publisher


def test_from_dict_composer_correct(setup_from_dict):
    result_rivendell_cart = setup_from_dict.result_rivendell_cart
    assert setup_from_dict.expected_composer == result_rivendell_cart.composer


def test_from_dict_conductor_correct(setup_from_dict):
    result_rivendell_cart = setup_from_dict.result_rivendell_cart
    assert setup_from_dict.expected_conductor == result_rivendell_cart.conductor


def test_from_dict_song_id_correct(setup_from_dict):
    result_rivendell_cart = setup_from_dict.result_rivendell_cart
    assert setup_from_dict.expected_song_id == result_rivendell_cart.song_id


def test_from_dict_user_defined_correct(setup_from_dict):
    result_rivendell_cart = setup_from_dict.result_rivendell_cart
    assert setup_from_dict.expected_user_defined == result_rivendell_cart.user_defined


def test_from_dict_description_correct(setup_from_dict):
    result_rivendell_cart = setup_from_dict.result_rivendell_cart
    assert setup_from_dict.expected_description == result_rivendell_cart.description


def test_from_dict_outcue_correct(setup_from_dict):
    result_rivendell_cart = setup_from_dict.result_rivendell_cart
    assert setup_from_dict.expected_outcue == result_rivendell_cart.outcue


def test_from_dict_filename_correct(setup_from_dict):
    result_rivendell_cart = setup_from_dict.result_rivendell_cart
    assert setup_from_dict.expected_filename == result_rivendell_cart.filename


def test_from_dict_length_correct(setup_from_dict):
    result_rivendell_cart = setup_from_dict.result_rivendell_cart
    assert setup_from_dict.expected_length == result_rivendell_cart.length


def test_from_dict_start_point_correct(setup_from_dict):
    result_rivendell_cart = setup_from_dict.result_rivendell_cart
    assert setup_from_dict.expected_start_point == result_rivendell_cart.start_point


def test_from_dict_end_point_correct(setup_from_dict):
    result_rivendell_cart = setup_from_dict.result_rivendell_cart
    assert setup_from_dict.expected_end_point == result_rivendell_cart.end_point


def test_from_dict_segue_start_point_correct(setup_from_dict):
    result_rivendell_cart = setup_from_dict.result_rivendell_cart
    assert setup_from_dict.expected_segue_start_point == result_rivendell_cart.segue_start_point


def test_from_dict_segue_end_point_correct(setup_from_dict):
    result_rivendell_cart = setup_from_dict.result_rivendell_cart
    assert setup_from_dict.expected_segue_end_point == result_rivendell_cart.segue_end_point


def test_from_dict_hook_start_point_correct(setup_from_dict):
    result_rivendell_cart = setup_from_dict.result_rivendell_cart
    assert setup_from_dict.expected_hook_start_point == result_rivendell_cart.hook_start_point


def test_from_dict_hook_end_point_correct(setup_from_dict):
    result_rivendell_cart = setup_from_dict.result_rivendell_cart
    assert setup_from_dict.expected_hook_end_point == result_rivendell_cart.hook_end_point


def test_from_dict_talk_start_point_correct(setup_from_dict):
    result_rivendell_cart = setup_from_dict.result_rivendell_cart
    assert setup_from_dict.expected_talk_start_point == result_rivendell_cart.talk_start_point


def test_from_dict_talk_end_point_correct(setup_from_dict):
    result_rivendell_cart = setup_from_dict.result_rivendell_cart
    assert setup_from_dict.expected_talk_end_point == result_rivendell_cart.talk_end_point


def test_from_dict_fadeup_point_correct(setup_from_dict):
    result_rivendell_cart = setup_from_dict.result_rivendell_cart
    assert setup_from_dict.expected_fadeup_point == result_rivendell_cart.fadeup_point


def test_from_dict_fadedown_point_correct(setup_from_dict):
    result_rivendell_cart = setup_from_dict.result_rivendell_cart
    assert setup_from_dict.expected_fadedown_point == result_rivendell_cart.fadedown_point


def test_from_dict_sched_codes_correct(setup_from_dict):
    result_rivendell_cart = setup_from_dict.result_rivendell_cart
    assert setup_from_dict.expected_sched_codes == result_rivendell_cart.sched_codes

def test_length_in_seconds_correct(setup_from_dict):
    result_rivendell_cart = setup_from_dict.result_rivendell_cart
    assert setup_from_dict.expected_length_in_seconds == result_rivendell_cart.length_in_seconds()

def test_length_in_seconds_longer_than_one_hour():
    source_dict = {
        "CART_NUMBER": '100340',
        "CUT_NUMBER": '1',
        "TYPE": "audio",
        "GROUP_NAME": '"ALTERNATIV"',
        "TITLE": '"Do You Want Love"',
        "ARTIST": '"Dave Depper"',
        "ALBUM": '""',
        "YEAR": '"2017"',
        "ISRC": '""',
        "ISCI": '""',
        "LABEL": '""',
        "CLIENT": '""',
        "AGENCY": '""',
        "PUBLISHER": '""',
        "COMPOSER": '"Imported from WOAFR: A17/0461"',
        "CONDUCTOR": '""',
        "SONG_ID": '""',
        "USER_DEFINED": '""',
        "DESCRIPTION": '"Do You Want Love"',
        "OUTCUE": '""',
        "FILENAME": '"100340_001.wav"',
        "LENGTH": '1:07:14',
        "START_POINT": '0',
        "END_POINT": '434000',
        "SEGUE_START_POINT": '432000',
        "SEGUE_END_POINT": '434000',
        "HOOK_START_POINT": '-1',
        "HOOK_END_POINT": '-1',
        "TALK_START_POINT": '0',
        "TALK_END_POINT": '40000',
        "FADEUP_POINT": '-1',
        "FADEDOWN_POINT": '-1',
        "SCHED_CODES": '"2015s"'
    }

    result_rivendell_cart = RivendellCart.from_dict(source_dict=source_dict)

    expected_length_in_seconds = 4034

    assert result_rivendell_cart.length_in_seconds() == expected_length_in_seconds



"""
@Author = 'Michael Stanley'

============ Change Log ============
2023-Jan-20 = Change license from GPLv2 to GPLv3.

2020-Oct-07 = Created.

============ License ============
Copyright (C) 2020, 2023, 2025 Michael Stanley

This file is part of wmul_rivendell.

wmul_rivendell is free software: you can redistribute it and/or modify it 
under the terms of the GNU General Public License as published by the Free 
Software Foundation, either version 3 of the License, or (at your option) any 
later version.

wmul_rivendell is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with 
wmul_rivendell. If not, see <https://www.gnu.org/licenses/>. 
"""
import cachetools
import datetime
import pathlib
import pytest
import wmul_test_utils

from wmul_rivendell import RivendellAudioImporter


@pytest.fixture(scope="function")
def setup__gather_file_names(mocker, caplog):
    mock_source_path_1 = "mock_source_path_1"
    mock_source_path_2 = "mock_source_path_2"
    mock_source_paths = [
        mock_source_path_1,
        mock_source_path_2
    ]
    mock_previously_seen_files = "mock_previously_seen_files"

    mock_previously_sent_to_importer_cache = "mock_previously_sent_to_importer_cache"

    mock_datetime_datetime_now = datetime.datetime(year=2020, month=6, day=30, hour=12, minute=5, second=10)

    mock_datetime_datetime_now_function = mocker.Mock(
        return_value=mock_datetime_datetime_now
    )
    mocker.patch("wmul_rivendell.RivendellAudioImporter.datetime.datetime",
                 now=mock_datetime_datetime_now_function)

    mock_files_seen_this_time_this_path_source_path_1 = {
        "mock_key_1": "mock_value_1",
        "mock_key_2": "mock_value_2"
    }
    mock_files_seen_this_time_this_path_source_path_2 = {
        "mock_key_3": "mock_value_3",
        "mock_key_4": "mock_value_4"
    }

    expected_files_seen_this_time = {
        **mock_files_seen_this_time_this_path_source_path_1,
        **mock_files_seen_this_time_this_path_source_path_2
    }

    mock_files_for_importer_this_path_source_path_1 = {
        "mock_file_1",
        "mock_file_2"
    }

    mock_files_for_importer_this_path_source_path_2 = {
        "mock_file_1",
        "mock_file_2"
    }

    expected_files_for_importer = [
        *mock_files_for_importer_this_path_source_path_1,
        *mock_files_for_importer_this_path_source_path_2
    ]

    def mock__gather_names_from_this_path_function(psf, dtn, sp, pstic):
        if sp == mock_source_path_1:
            return mock_files_seen_this_time_this_path_source_path_1, mock_files_for_importer_this_path_source_path_1
        elif sp == mock_source_path_2:
            return mock_files_seen_this_time_this_path_source_path_2, mock_files_for_importer_this_path_source_path_2

    mock__gather_names_from_this_path = mocker.Mock(
        side_effect= mock__gather_names_from_this_path_function
    )

    mocker.patch("wmul_rivendell.RivendellAudioImporter._gather_names_from_this_path",
                 mock__gather_names_from_this_path)

    result_files_seen_this_time, result_files_for_importer = \
        RivendellAudioImporter._gather_file_names(
            mock_source_paths,
            mock_previously_seen_files,
            mock_previously_sent_to_importer_cache
        )

    expected_calls = [
        mocker.call(mock_previously_seen_files, mock_datetime_datetime_now, mock_source_paths[0],
                    mock_previously_sent_to_importer_cache),

        mocker.call(mock_previously_seen_files, mock_datetime_datetime_now, mock_source_paths[1],
                    mock_previously_sent_to_importer_cache),
    ]

    expected_log_messages = [
        f"With Source Paths: {mock_source_paths}\t Previously Seen Files: {mock_previously_seen_files}\t and "
        f"{mock_previously_sent_to_importer_cache}",

        f"Datetime resolved: {mock_datetime_datetime_now}",
        f"Working on {mock_source_path_1}",
        f"Working on {mock_source_path_2}",

        f"Returning Files_seen_this_time: {expected_files_seen_this_time} and "
        f"Files_for_importer: {expected_files_for_importer}"
    ]

    return wmul_test_utils.make_namedtuple(
        "setup__gather_file_names",
        mock_datetime_datetime_now_function=mock_datetime_datetime_now_function,
        expected_files_seen_this_time=expected_files_seen_this_time,
        expected_files_for_importer=expected_files_for_importer,
        mock__gather_names_from_this_path=mock__gather_names_from_this_path,
        result_files_seen_this_time=result_files_seen_this_time,
        result_files_for_importer=result_files_for_importer,
        expected_calls=expected_calls,
        expected_log_messages=expected_log_messages,
        result_log_messages=caplog.messages
    )


def test_datetime_datetime_now_called_correctly(setup__gather_file_names):
    setup__gather_file_names.mock_datetime_datetime_now_function.assert_called_once_with()


def test__gather_names_called_correctly(setup__gather_file_names):
    setup__gather_file_names.mock__gather_names_from_this_path.assert_has_calls(
        setup__gather_file_names.expected_calls)

    assert setup__gather_file_names.mock__gather_names_from_this_path.call_count == \
           len(setup__gather_file_names.expected_calls)


def test_files_seen_this_time_correct(setup__gather_file_names):
    assert setup__gather_file_names.expected_files_seen_this_time == \
           setup__gather_file_names.result_files_seen_this_time


def test_files_for_importer_correct(setup__gather_file_names):
    assert sorted(setup__gather_file_names.expected_files_for_importer) == \
           sorted(setup__gather_file_names.result_files_for_importer)


def test_log_messages_correct(setup__gather_file_names):
    assert sorted(setup__gather_file_names.expected_log_messages) == \
           sorted(setup__gather_file_names.result_log_messages)


def test__gather_file_names_end_to_end(fs, mocker):
    source_path_1 = pathlib.Path(r"\source\group1")
    source_path_2 = pathlib.Path(r"\source\group2")

    fs.create_dir(source_path_1)
    fs.create_dir(source_path_2)

    file_1_1_path = source_path_1 / "file1.wav"
    file_1_2_path = source_path_1 / "file2.wav"
    file_1_3_path = source_path_1 / "file3.wav"
    file_2_3_path = source_path_2 / "file3.wav"
    file_2_4_path = source_path_2 / "file4.wav"
    file_2_5_path = source_path_2 / "file5.wav"

    initial_size = 1_000
    intermediate_size = 5_000
    final_size = 10_000

    file_1_1_fake_file = fs.create_file(file_1_1_path, st_size=final_size)
    file_1_2_fake_file = fs.create_file(file_1_2_path, st_size=initial_size)
    file_1_3_fake_file = fs.create_file(file_1_3_path, st_size=intermediate_size)
    file_2_3_fake_file = fs.create_file(file_2_3_path, st_size=intermediate_size)
    file_2_4_fake_file = fs.create_file(file_2_4_path, st_size=intermediate_size)
    file_2_5_fake_file = fs.create_file(file_2_5_path, st_size=initial_size)

    first_pass_mock_datetime_datetime_now = datetime.datetime(year=2020, month=6, day=30, hour=12, minute=5, second=10)
    second_pass_mock_datetime_datetime_now = datetime.datetime(year=2020, month=6, day=30, hour=12, minute=5, second=30)
    third_pass_mock_datetime_datetime_now = datetime.datetime(year=2020, month=6, day=30, hour=12, minute=5, second=50)
    fourth_pass_mock_datetime_datetime_now = datetime.datetime(year=2020, month=6, day=30, hour=12, minute=6, second=10)
    fifth_pass_mock_datetime_datetime_now = datetime.datetime(year=2020, month=6, day=30, hour=12, minute=6, second=30)
    sixth_pass_mock_datetime_datetime_now = datetime.datetime(year=2020, month=6, day=30, hour=12, minute=6, second=50)

    patched_datetime_datetime = mocker.patch("wmul_rivendell.RivendellAudioImporter.datetime.datetime",
                                             now=lambda: first_pass_mock_datetime_datetime_now)

    source_paths = [source_path_1, source_path_2]
    previously_seen_files = {}
    previously_sent_to_importer_cache = cachetools.TTLCache(maxsize=10_000, ttl=180, timer=lambda: 1)

    # First pass. Should see all six files in files_seen_this_time and no files in files_for_importer.

    result_files_seen_this_time, result_files_for_importer = \
        RivendellAudioImporter._gather_file_names(
            source_paths=source_paths,
            previously_seen_files=previously_seen_files,
            previously_sent_to_importer_cache=previously_sent_to_importer_cache
        )

    first_pass_expected_files_seen_this_time_file_paths = [
        file_1_1_path,
        file_1_2_path,
        file_1_3_path,
        file_2_3_path,
        file_2_4_path,
        file_2_5_path
    ]

    first_pass_expected_files_for_importer = []

    assert sorted(result_files_seen_this_time.keys()) == \
           sorted(first_pass_expected_files_seen_this_time_file_paths)
    assert sorted(result_files_for_importer) == sorted(first_pass_expected_files_for_importer)

    # Second pass. Should still see all six files in files_seen_this_time and no files in files_for_importer.

    file_1_2_fake_file.st_size = intermediate_size
    file_1_3_fake_file.st_size = final_size
    file_2_3_fake_file.st_size = final_size
    file_2_4_fake_file.st_size = final_size
    file_2_5_fake_file.st_size = intermediate_size

    patched_datetime_datetime.now = lambda: second_pass_mock_datetime_datetime_now

    result_files_seen_this_time, result_files_for_importer = \
        RivendellAudioImporter._gather_file_names(
            source_paths=source_paths,
            previously_seen_files=result_files_seen_this_time,
            previously_sent_to_importer_cache=previously_sent_to_importer_cache
        )

    second_pass_expected_files_seen_this_time_file_paths = [
        file_1_1_path,
        file_1_2_path,
        file_1_3_path,
        file_2_3_path,
        file_2_4_path,
        file_2_5_path
    ]

    second_pass_expected_files_for_importer = []

    assert sorted(result_files_seen_this_time.keys()) == \
           sorted(second_pass_expected_files_seen_this_time_file_paths)
    assert sorted(result_files_for_importer) == sorted(second_pass_expected_files_for_importer)

    # Third pass. Should see all but file_1_1 seen_this_time and should see file_1_1 for importer.

    file_1_2_fake_file.st_size = final_size
    file_2_5_fake_file.st_size = final_size

    patched_datetime_datetime.now = lambda: third_pass_mock_datetime_datetime_now

    result_files_seen_this_time, result_files_for_importer = \
        RivendellAudioImporter._gather_file_names(
            source_paths=source_paths,
            previously_seen_files=result_files_seen_this_time,
            previously_sent_to_importer_cache=previously_sent_to_importer_cache
        )

    result_files_for_importer_file_paths = [file_item.file_path for file_item in result_files_for_importer]

    third_pass_expected_files_seen_this_time_file_paths = [
        file_1_2_path,
        file_1_3_path,
        file_2_3_path,
        file_2_4_path,
        file_2_5_path
    ]

    third_pass_expected_files_for_importer_file_paths = [file_1_1_path]

    assert sorted(result_files_seen_this_time.keys()) == \
           sorted(third_pass_expected_files_seen_this_time_file_paths)
    assert sorted(result_files_for_importer_file_paths) == \
           sorted(third_pass_expected_files_for_importer_file_paths)

    # Fourth pass. Should see 1_2 and 2_5 seen_this_time. Should see file_1_3, 2_3, and 2_4 for importer.

    previously_sent_to_importer_cache[file_1_1_path] = True

    patched_datetime_datetime.now = lambda: fourth_pass_mock_datetime_datetime_now

    result_files_seen_this_time, result_files_for_importer = \
        RivendellAudioImporter._gather_file_names(
            source_paths=source_paths,
            previously_seen_files=result_files_seen_this_time,
            previously_sent_to_importer_cache=previously_sent_to_importer_cache
        )

    result_files_for_importer_file_paths = [file_item.file_path for file_item in result_files_for_importer]

    fourth_pass_expected_files_seen_this_time_file_paths = [
        file_1_2_path,
        file_2_5_path
    ]

    fourth_pass_expected_files_for_importer_file_paths = [
        file_1_3_path,
        file_2_3_path,
        file_2_4_path
    ]

    assert sorted(result_files_seen_this_time.keys()) == \
           sorted(fourth_pass_expected_files_seen_this_time_file_paths)
    assert sorted(result_files_for_importer_file_paths) == \
           sorted(fourth_pass_expected_files_for_importer_file_paths)

    previously_sent_to_importer_cache[file_1_3_path] = True
    previously_sent_to_importer_cache[file_2_3_path] = True
    previously_sent_to_importer_cache[file_2_4_path] = True

    # Fifth pass. Should not see any seen_this_time Should see 1_2 and 2_5 to importer.

    patched_datetime_datetime.now = lambda: fifth_pass_mock_datetime_datetime_now

    result_files_seen_this_time, result_files_for_importer = \
        RivendellAudioImporter._gather_file_names(
            source_paths=source_paths,
            previously_seen_files=result_files_seen_this_time,
            previously_sent_to_importer_cache=previously_sent_to_importer_cache
        )

    result_files_for_importer_file_paths = [file_item.file_path for file_item in result_files_for_importer]

    fifth_pass_expected_files_seen_this_time_file_paths = []

    fifth_pass_expected_files_for_importer_file_paths = [
        file_1_2_path,
        file_2_5_path
    ]

    assert sorted(result_files_seen_this_time.keys()) == \
           sorted(fifth_pass_expected_files_seen_this_time_file_paths)
    assert sorted(result_files_for_importer_file_paths) == \
           sorted(fifth_pass_expected_files_for_importer_file_paths)

    previously_sent_to_importer_cache[file_1_2_path] = True
    previously_sent_to_importer_cache[file_2_5_path] = True

    # Sixth pass. Should not see any seen_this_time or files_for_importer.

    patched_datetime_datetime.now = lambda: sixth_pass_mock_datetime_datetime_now

    result_files_seen_this_time, result_files_for_importer = \
        RivendellAudioImporter._gather_file_names(
            source_paths=source_paths,
            previously_seen_files=result_files_seen_this_time,
            previously_sent_to_importer_cache=previously_sent_to_importer_cache
        )

    result_files_for_importer_file_paths = [file_item.file_path for file_item in result_files_for_importer]

    sixth_pass_expected_files_seen_this_time_file_paths = []
    sixth_pass_expected_files_for_importer_file_paths = []

    assert sorted(result_files_seen_this_time.keys()) == \
           sorted(sixth_pass_expected_files_seen_this_time_file_paths)
    assert sorted(result_files_for_importer_file_paths) == \
           sorted(sixth_pass_expected_files_for_importer_file_paths)




"""
@Author = 'Michael Stanley'

This script takes the "Cart Data Dump (CSV)" from RD Library and generates statistics about each group.

============ Change Log ============
2025-Jan-30 = Add logic to rename the DatabaseStatistics.output_filename if that file already exists.

2025-Jan-29 = Remove to_list_for_csv and get_header_list, they are no longer needed. CSVs are generated using pandas 
              now.

              Flip the orientation of the statistics_limit sheet when outputting to an excel file. To make it match 
              the orientation of the csv file.

              Extract the identical lines for the creation of the pandas data from _write_csv and _write_excel to a 
              single method.

              Add option to output the full set of statistics, or just the number of songs and the boundaries.

2025-Jan-23 = Make the upper and lower bound mulitples variables instead of hard coded.

              Add logic to optionally write the statistics limits values to the output file.

              Rewrite _write_csv to use pandas instead of direct csv. Will eliminate the need for a separate 
              to_list_for_csv method in RivendellGroupStatistics.

2025-Jan-22 = When the stdev is less than 15 seconds, don't bother calculating the exclusion limits. There is too 
              little variance for those calculations to be meaningful and correct.

              Refactor the _remove_outliers method.

              Add ability to write to an Excel spreadsheet.

              Add column to calculate the percentage of songs that are excluded.

              Add logic to output either an excel file or a csv, depending on the suffix of the output_filename.

              Extract _remove_outliers from a static method on RivendellGroupStatistics to a standalone.

              Make smallest_stdev and minimum_population into variables instead of hard-coded.

2025-Jan-21 = No longer try to remove outliers if the standard deviation of the population is less than 15 seconds.

2025-Jan-17 = Extract statistics, expand generated data.

2025-Jan-03 = Created.

============ License ============
Copyright (C) 2025 Michael Stanley

This file is part of wmul_rivendell.

wmul_rivendell is free software: you can redistribute it and/or modify it 
under the terms of the GNU General Public License as published by the Free 
Software Foundation, either version 3 of the License, or (at your option) any 
later version.

wmul_rivendell is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with 
wmul_rivendell. If not, see <https://www.gnu.org/licenses/>. 
"""
import math
import numpy as np
import pandas as pd
from collections import defaultdict
from dataclasses import dataclass
from datetime import timedelta
from pathlib import Path

import wmul_logger

_logger = wmul_logger.get_logger()

_MAX_TIME = 86_399


@dataclass
class StatisticsLimits:
    smallest_stdev: int = 15
    minimum_population_for_outliers: int = 4
    lower_bound_multiple: float = 1.5
    upper_bound_multiple: float = 3.0

    def to_pandas_series(self):
        return pd.Series(
            {
                "Smallest Standard Deviation": str(timedelta(seconds=int(self.smallest_stdev))),
                "Minimum Population for Outliers": self.minimum_population_for_outliers,
                "Lower Bound Multiple": self.lower_bound_multiple,
                "Upper Bound Multiple": self.upper_bound_multiple
            }
        )
    

class RivendellGroupStatistics:

    def __init__(self, group_name: str, songs_in_group: list, stats_limits: StatisticsLimits):
        self.group_name = group_name
        self.stats_limits = stats_limits
        times_of_this_group = np.array([this_item.length_in_seconds() for this_item in songs_in_group])
        times_of_this_group.sort()
        
        self.number_of_songs = times_of_this_group.size
        self.shortest_song_length = times_of_this_group.min()
        self.longest_song_length = times_of_this_group.max()

        outliers_exluded, outlier_lower_limit, outlier_upper_limit = _remove_outliers(
            times_of_this_group=times_of_this_group, 
            stats_limits=stats_limits
        )

        self.outlier_limits = (round(outlier_lower_limit), round(outlier_upper_limit))
        self.mean = round(outliers_exluded.mean())
        
        stdev = outliers_exluded.std()
        if math.isnan(stdev):
            self.stdev = 0
        else:
            self.stdev = round(stdev)

        if stdev > self.stats_limits.smallest_stdev:
            lower_bound = self.mean - (stats_limits.lower_bound_multiple * stdev)
            if lower_bound < 0:
                self.lower_bound = 0
            else:
                self.lower_bound = RivendellGroupStatistics._nearest_15(lower_bound)
            upper_bound = self.mean + (stats_limits.upper_bound_multiple * stdev)
            self.upper_bound = RivendellGroupStatistics._nearest_15(upper_bound)

            shorter_than_lower_bound = [
                song_length for song_length in times_of_this_group if song_length < self.lower_bound
            ]
            self.number_of_songs_shorter_than_lower_bound = len(shorter_than_lower_bound)

            longer_than_upper_bound = [
                song_length for song_length in times_of_this_group if song_length > self.upper_bound
            ]
            self.number_of_songs_longer_than_upper_bound = len(longer_than_upper_bound)
            total_songs_excluded = self.number_of_songs_shorter_than_lower_bound + \
                self.number_of_songs_longer_than_upper_bound
            percentage_of_songs_excluded = (total_songs_excluded / self.number_of_songs) * 100
            percentage_of_songs_excluded = round(percentage_of_songs_excluded, 1)
            self.percentage_of_songs_excluded = percentage_of_songs_excluded
        else:
            # If STDev is below stats_limits.smallest_stdev, there is not enough variance in the lengths of the songs 
            # for the exclusion to be meaningful and correct.
            self.lower_bound = 0
            self.number_of_songs_shorter_than_lower_bound = 0
            self.upper_bound = _MAX_TIME
            self.number_of_songs_longer_than_upper_bound = 0
            self.percentage_of_songs_excluded = 0
    
    @staticmethod
    def _nearest_15(input_number):
        input_number = round(input_number)
        mod_15 = input_number % 15
        if mod_15 == 0:
            return input_number
        elif mod_15 < 8:
            return input_number - mod_15
        else:
            return input_number + (15 - mod_15)
        
    def to_pandas_series(self, write_full_statistics):
        lower_outlier_limit, upper_outlier_limit = self.outlier_limits
        lower_outlier_limit = str(timedelta(seconds=int(lower_outlier_limit)))
        upper_outlier_limit = str(timedelta(seconds=int(upper_outlier_limit)))

        statistics = {
            "Number of Songs": self.number_of_songs,
            "Lower Bound": str(timedelta(seconds=int(self.lower_bound))),
            "Upper Bound": str(timedelta(seconds=int(self.upper_bound)))
        }

        if write_full_statistics:
            # Construct a new dict, rather than updating the exising one, so that the new keys and old keys can be 
            # interleaved.
            statistics = {
                "Number of Songs": statistics["Number of Songs"],
                "Shortest Song Length": str(timedelta(seconds=int(self.shortest_song_length))),
                "Longest Song Length": str(timedelta(seconds=int(self.longest_song_length))),
                "Outlier Limits": (lower_outlier_limit, upper_outlier_limit),
                "Mean": str(timedelta(seconds=int(self.mean))),
                "Standard Deviation": str(timedelta(seconds=int(self.stdev))),
                "Lower Bound": statistics["Lower Bound"],
                "Number of Songs < Lower Bound": self.number_of_songs_shorter_than_lower_bound,
                "Upper Bound": statistics["Upper Bound"],
                "Number of Songs > Upper Bound": self.number_of_songs_longer_than_upper_bound,
                "Percent of Songs Excluded": self.percentage_of_songs_excluded
            }

        return pd.Series(statistics)


def _remove_outliers(times_of_this_group: np.array, stats_limits: StatisticsLimits):
    if ((times_of_this_group.size > stats_limits.minimum_population_for_outliers) and 
        (times_of_this_group.std() >= stats_limits.smallest_stdev)):
        q25, q75 = np.percentile(times_of_this_group, [25, 75])
        iqr = q75 - q25
        iqr_times_1_point_5 = iqr * 1.5
        lower_limit = q25 - iqr_times_1_point_5
        upper_limit = q75 + iqr_times_1_point_5
        outliers_excluded = np.array([this_item for this_item in times_of_this_group if lower_limit < this_item < upper_limit])
        outliers_excluded.sort()
        return outliers_excluded, lower_limit, upper_limit
    else:
        return times_of_this_group, 0, _MAX_TIME


@dataclass
class DatabaseStatistics:
    rivendell_carts: list
    output_filename: Path
    stats_limits: StatisticsLimits
    write_limits: bool
    write_full_statistics: bool

    def _organize_by_rivendell_group(self, unorganized_carts):
        organized_by_rivendell_group = defaultdict(list)

        for rivendell_cart in unorganized_carts:
            organized_by_rivendell_group[rivendell_cart.group_name].append(rivendell_cart)
        
        return organized_by_rivendell_group
    
    def _calculate_statistics_per_group(self, organized_carts):
        statistics_per_group = dict()
        for group_name, songs_in_group in organized_carts.items():
            _logger.debug(f"Working on {group_name}")
            statistics_per_group[group_name] = RivendellGroupStatistics(
                group_name=group_name, 
                songs_in_group=songs_in_group,
                stats_limits=self.stats_limits
            )
        return statistics_per_group

    def _generate_pandas_data(self, statistics_per_group):
        df_limits = pd.DataFrame(
            { "Statistics Limits": self.stats_limits.to_pandas_series() }
        )
        df_limits = df_limits.T

        group_names = sorted(statistics_per_group.keys())
        df_data = pd.DataFrame(
            { group_name: statistics_per_group[group_name].to_pandas_series(self.write_full_statistics) 
              for group_name in group_names }
        )
        df_data = df_data.T

        return df_limits, df_data

    def _write_csv(self, statistics_per_group):
        df_limits, df_data = self._generate_pandas_data(statistics_per_group=statistics_per_group)
        with open(str(self.output_filename), newline="", mode="wt", errors="replace") as statistics_output:
            if self.write_limits:
                df_limits.to_csv(statistics_output)
            df_data.to_csv(statistics_output, index_label="Group Name")

    def _write_excel(self, statistics_per_group):
        df_limits, df_data = self._generate_pandas_data(statistics_per_group=statistics_per_group)
        with pd.ExcelWriter(self.output_filename) as writer:
            if self.write_limits:
                df_limits.to_excel(writer, sheet_name="Limits")
            df_data.to_excel(writer, sheet_name="Data")

    def run_script(self):
        _logger.debug(f"Starting DatabaseStatistics.run_script()")
        organized_by_rivendell_group = self._organize_by_rivendell_group(unorganized_carts=self.rivendell_carts)
        statistics_per_group = self._calculate_statistics_per_group(organized_carts=organized_by_rivendell_group)
        output_filename = self.output_filename
        if output_filename.exists():
            new_filename = (output_filename.parent / 
                (output_filename.stem + "_old" + output_filename.suffix))
            output_filename.rename(new_filename)
        if output_filename.suffix == '.xlsx':
            self._write_excel(statistics_per_group=statistics_per_group)
        else:
            self._write_csv(statistics_per_group=statistics_per_group)



"""
@Author = 'Michael Stanley'

============ Change Log ============
2023-Jan-20 = Change license from GPLv2 to GPLv3.

2020-Aug-03 = Created.

============ License ============
Copyright (C) 2020, 2023 Michael Stanley

This file is part of wmul_rivendell.

wmul_rivendell is free software: you can redistribute it and/or modify it 
under the terms of the GNU General Public License as published by the Free 
Software Foundation, either version 3 of the License, or (at your option) any 
later version.

wmul_rivendell is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with 
wmul_rivendell. If not, see <https://www.gnu.org/licenses/>. 
"""



"""
@Author = 'Michael Stanley'

============ Change Log ============
2025-Jan-03 = Add two additional checks. The first checks to see if generate_importer_command returned None. 
              If so, then the file has no group. (It is in the root folder.) The second checks the result from 
              rdimport to see if it failed to open the file.

2023-Feb-27 = Catch crashes that occur when renaming a bad file and the file 
              is locked by another user.

2023-Jan-20 = Back-insert items into change log.
              Change license from GPLv2 to GPLv3.

2022-Jun-09 = Improve the crash handling when the problem is a network burp.

2022-May-16 = Fix bug in subprocess due to back-porting behind Python 3.7. 
              Python 3.6 does not have the subprocess.run capture_output param.

2022-May-05 = Make .wav case-insensitive.

2021-May-26 = Added recovery from certain network burps.

2020-Oct-15 = Added ability to tell rdimport where to log.
              Now renamed files as _FAILED when rdimport rejects them.
              Improve own logging.

2020-Jun-30 = Created.

============ License ============
Copyright (C) 2020-2023, 2025 Michael Stanley

This file is part of wmul_rivendell.

wmul_rivendell is free software: you can redistribute it and/or modify it 
under the terms of the GNU General Public License as published by the Free 
Software Foundation, either version 3 of the License, or (at your option) any 
later version.

wmul_rivendell is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with 
wmul_rivendell. If not, see <https://www.gnu.org/licenses/>. 
"""
from dataclasses import dataclass
from functools import total_ordering
from pathlib import Path
from time import sleep
import datetime
import cachetools
import subprocess

import wmul_logger

_logger = wmul_logger.get_logger()


@total_ordering
class FileInformation:

    def __init__(self, file_path: Path, timestamp: datetime.datetime, source_path: Path):
        self.file_path = file_path
        self.file_size = file_path.stat().st_size
        self.timestamp = timestamp
        self.relative_path = file_path.relative_to(source_path)

    def __str__(self):
        return f"Path: {self.file_path}, Size: {self.file_size}, Timestamp: {self.timestamp}, " \
               f"Relative_Path: {self.relative_path}."

    def __eq__(self, other):
        return str(self) == str(other)

    def __lt__(self, other):
        return str(self) < str(other)

    def __hash__(self):
        return hash(str(self))

    def same_size_as(self, other):
        return self.file_size == other.file_size

    def new_info_much_newer(self, second_file, minimum_difference):
        difference = self.timestamp - second_file.timestamp
        return difference < datetime.timedelta(seconds=-minimum_difference)

    def generate_importer_command(self, log_argument):
        _logger.debug(f"With {self}")
        relative_path = self.relative_path
        relative_path_parts = relative_path.parts
        len_of_relative_path_parts = len(relative_path_parts)
        if len_of_relative_path_parts == 1:
            _logger.debug(f"This file does not have a group. {str(self.file_path)}")
            return None
        elif len_of_relative_path_parts == 2:
            _logger.debug("No scheduler codes.")
            group_folder, file_path = relative_path_parts
            scheduler_codes_options = []
        else:
            _logger.debug("Has scheduler codes.")
            group_folder, *scheduler_code_folders, file_path = relative_path_parts
            scheduler_codes_options = [f'--add-scheduler-code={scheduler_code}'
                                       for scheduler_code in scheduler_code_folders]

        import_command = [
            *scheduler_codes_options,
            "--autotrim-level=0",
            "--normalization-level=0",
            "--title-from-cartchunk-cutid",
            "--delete-source",
            f'--set-string-description="{self.file_path.name}"',
            "--verbose"
        ]

        if log_argument:
            import_command.append(log_argument)

        import_command.extend(
            [group_folder, str(self.file_path)]
        )
        return import_command

    def _failed(self, new_suffix):
        old_suffix = self.file_path.suffix
        new_path = self.file_path.with_suffix(old_suffix + new_suffix)
        try: 
            self.file_path.rename(new_path)
        except OSError as ose:
            _logger.debug(ose)
            if ose.errno == 16: # Device or Resource Busy (Someone else has the file locked).
                return
            else:
                raise

    def failed_rdimport(self):
        self._failed("_RDIMPORT_FAILED")

    def failed_group(self):
        self._failed("_NO_GROUP_FAILED")


@dataclass
class ImportRivendellFileWithFileSystemMetadataArguments:
    source_paths: list
    cache_duration: int
    rdimport_syslog: bool
    rdimport_log_file_name: str


def _gather_file_names(source_paths, previously_seen_files, previously_sent_to_importer_cache):
    _logger.debug(f"With Source Paths: {source_paths}\t Previously Seen Files: {previously_seen_files}\t "
                  f"and {previously_sent_to_importer_cache}")
    files_seen_this_time = {}
    files_for_importer = []
    datetime_now = datetime.datetime.now()
    _logger.debug(f"Datetime resolved: {datetime_now}")
    for source_path in source_paths:
        _logger.debug(f"Working on {source_path}")
        files_seen_this_time_this_path, files_for_importer_this_path = \
            _gather_names_from_this_path(previously_seen_files, datetime_now, source_path,
                                         previously_sent_to_importer_cache)
        files_seen_this_time.update(files_seen_this_time_this_path)
        files_for_importer.extend(files_for_importer_this_path)
    _logger.debug(f"Returning Files_seen_this_time: {files_seen_this_time} and Files_for_importer: {files_for_importer}")
    return files_seen_this_time, files_for_importer


def _gather_names_from_this_path(previously_seen_files, datetime_now, source_path, previously_sent_to_importer_cache):
    _logger.debug(
        f"With Previously Seen Files: {previously_seen_files}, Datetime_now: {datetime_now}, "
        f"Source Path: {source_path}, and Previously Sent to Importer Cache: {previously_sent_to_importer_cache}"
    )
    matching_file_paths = source_path.rglob("*.[Ww][Aa][Vv]")
    files_seen_this_time = {}
    files_for_importer = []
    for file_path in matching_file_paths:
        _logger.debug(f"Working on {file_path}")
        if file_path not in previously_sent_to_importer_cache:
            _logger.debug(f"Have not sent {file_path} to the importer in the previous "
                          f"{previously_sent_to_importer_cache.ttl} seconds")
            seen_this_time, for_importer = _check_this_file(previously_seen_files, file_path, datetime_now, source_path)
            if seen_this_time:
                _logger.debug(f"Seen this time! {file_path}")
                files_seen_this_time[file_path] = seen_this_time
            elif for_importer:
                _logger.info(f"For importer! {file_path}")
                files_for_importer.append(for_importer)
            else:
                _logger.debug(f"Not for seen this time or importer! {file_path}")
        else:
            _logger.debug(f"Sent a file with this name to the importer within the previous "
                          f"{previously_sent_to_importer_cache.ttl} seconds! {file_path}")
    _logger.debug(f"Returning files_seen_this_time: {files_seen_this_time} and files_for_importer {files_for_importer}")
    return files_seen_this_time, files_for_importer


def _check_this_file(previously_seen_files, file_path, datetime_now, source_path):
    _logger.debug(f"With {previously_seen_files}, {file_path}, {datetime_now}, {source_path}")
    if file_path.is_file():
        _logger.debug("Is File!")
        this_version = FileInformation(file_path=file_path, timestamp=datetime_now, source_path=source_path)
        previously_seen_version = previously_seen_files.get(file_path)
        _logger.debug(f"Previously seen version: {previously_seen_version}")
        if _previous_seen_and_same_size(previously_seen_version, this_version):
            _logger.debug("Previously seen and same size!")
            return _for_importer_or_seen_this_time(previously_seen_version, this_version)
        else:
            _logger.debug("New file or new size!")
            return this_version, None
    else:
        _logger.debug("Is not a file!")
        return None, None


def _for_importer_or_seen_this_time(previously_seen_version, this_version):
    _logger.debug(f"With {previously_seen_version}, {this_version}")
    if previously_seen_version.new_info_much_newer(this_version, 30):
        _logger.debug("This one is much newer!")
        return None, this_version
    else:
        _logger.debug("This one is not new enough!")
        return previously_seen_version, None


def _previous_seen_and_same_size(previously_seen_version, new_version):
    _logger.debug(f"With {previously_seen_version}, and {new_version}")
    return previously_seen_version and previously_seen_version.same_size_as(new_version)


def _run_importer_on_selected_files(files_for_importer, previously_sent_to_importer_cache, log_argument):
    for file_for_importer in files_for_importer:
        _logger.info(f"Sending to the importer: {file_for_importer.file_path}")
        importer_command = file_for_importer.generate_importer_command(log_argument)
        _logger.info(f"Importer command: {importer_command}")
        if not importer_command:
            file_for_importer.failed_group()
        previously_sent_to_importer_cache[file_for_importer.file_path] = True
        results = subprocess.run(["rdimport", *importer_command], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if results.returncode == 0:
            if ((b"is not readable or not a recognized format, skipping..." in results.stdout) or 
                (b"Unable to open " in results.stdout)):
                _logger.error(f"Error on {importer_command[-1]}, {results.args}, {results.stderr}, {results.stdout}")
                file_for_importer.failed_rdimport()
            else:
                _logger.debug(f"Success! {importer_command[-1]}, {results.args}, {results.stderr}, {results.stdout}")
        else:
            _logger.error(f"Error on {importer_command[-1]}, {results.args}, {results.stderr}, {results.stdout}")
            file_for_importer.failed_rdimport()


def run_script(arguments):
    _logger.debug(f"With {arguments}")
    previously_seen_files = {}
    previously_sent_to_importer_cache = cachetools.TTLCache(10_000, arguments.cache_duration)

    if arguments.rdimport_syslog:
        log_argument = "--log-syslog"
    elif arguments.rdimport_log_file_name:
        log_argument = f"--log-filename={arguments.rdimport_log_file_name}"
    else:
        log_argument = None

    _logger.info(f"With log_argument: {log_argument}")

    while True:
        exception_cache = cachetools.TTLCache(5, 3600)
        try:
            _logger.debug("Loop!")
            previously_seen_files, files_for_importer = _gather_file_names(
                arguments.source_paths,
                previously_seen_files,
                previously_sent_to_importer_cache
            )
            _run_importer_on_selected_files(files_for_importer, previously_sent_to_importer_cache, log_argument)
            sleep(10)
        except BlockingIOError as bio:
            _logger.debug(bio)
            exception_cache[datetime.datetime.now()] = True
            if exception_cache.currsize > 3:
                raise
            sleep(300)
        except OSError as ose:
            _logger.debug(ose)
            if ose.errno == 112:
                exception_cache[datetime.datetime.now()] = True
                if exception_cache.currsize > 3:
                    raise
                sleep(300)
            else:
                raise



"""
@Author = 'Michael Stanley'

============ Change Log ============
2025-Jan-03 = Created. 

============ License ============
Copyright (C) 2025 Michael Stanley

This file is part of wmul_rivendell.

wmul_rivendell is free software: you can redistribute it and/or modify it 
under the terms of the GNU General Public License as published by the Free 
Software Foundation, either version 3 of the License, or (at your option) any 
later version.

wmul_rivendell is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with 
wmul_rivendell. If not, see <https://www.gnu.org/licenses/>. 
"""
import pandas as pd
import pytest
from pathlib import Path

from wmul_rivendell.LoadCartDataDump import RivendellCart, CartType
from wmul_rivendell.DatabaseStatistics import DatabaseStatistics, StatisticsLimits

from wmul_test_utils import generate_true_false_matrix_from_list_of_strings, make_namedtuple

def test__organize_by_rivendell_group():
    defined_carts = [
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='LAUDANTIUM', title='facilis porro numquam nulla', artist='nostrum illum', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='LAUDANTIUM', title='illo deserunt inventore accusamus', artist='a quas', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='LAUDANTIUM', title='nihil ratione optio saepe', artist='quisquam fugit', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='LAUDANTIUM', title='praesentium nihil aperiam deleniti', artist='similique at', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='LAUDANTIUM', title='sapiente laboriosam aliquam impedit', artist='harum dolores', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='LAUDANTIUM', title='commodi voluptatibus nisi officia', artist='odit facere', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='LAUDANTIUM', title='occaecati dolorum eius molestias', artist='atque provident', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='LAUDANTIUM', title='odio quasi recusandae inventore', artist='autem at', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='LAUDANTIUM', title='accusamus consequatur quis quisquam', artist='quo ex', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='LAUDANTIUM', title='modi dolores aliquam aut', artist='exercitationem deserunt', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='ASPERIORES', title='reiciendis dolore consectetur harum', artist='dolore temporibus', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='ASPERIORES', title='labore commodi voluptatum nam', artist='perferendis perspiciatis', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='ASPERIORES', title='dolore quidem nulla culpa', artist='quas facilis', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='ASPERIORES', title='impedit rerum debitis magni', artist='deserunt vitae', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='ASPERIORES', title='rem id cupiditate culpa', artist='cum aliquid', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='ASPERIORES', title='dolor aliquam deserunt laboriosam', artist='ipsa dicta', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='ASPERIORES', title='blanditiis illum cupiditate tempora', artist='labore voluptatum', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='ASPERIORES', title='assumenda blanditiis natus incidunt', artist='error quibusdam', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='ASPERIORES', title='vero a eius ab', artist='aut quae', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='ASPERIORES', title='labore explicabo voluptas nisi', artist='dignissimos veritatis', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='magnam sed repellat iusto', artist='libero beatae', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='laborum laborum ipsum corrupti', artist='voluptatum nisi', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='alias dolorum qui odit', artist='recusandae perspiciatis', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='reiciendis architecto cupiditate aperiam', artist='optio nesciunt', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='omnis voluptas voluptates molestias', artist='similique natus', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='molestias non laborum eligendi', artist='eaque aliquid', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='placeat similique eligendi nobis', artist='expedita est', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='quasi consectetur sequi ab', artist='saepe cumque', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='natus doloremque earum repellendus', artist='deserunt corporis', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='assumenda aliquam itaque doloremque', artist='placeat ipsa', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='iure ad consequatur similique', artist='necessitatibus iusto', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='consectetur ad provident iusto', artist='cumque error', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='quod praesentium odit animi', artist='accusamus dolorum', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='in ex tempora aspernatur', artist='laborum tempora', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='facilis itaque dolore tenetur', artist='earum occaecati', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='reprehenderit alias hic necessitatibus', artist='maxime fugiat', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='modi recusandae atque provident', artist='necessitatibus consequatur', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='delectus itaque rerum recusandae', artist='minima neque', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='voluptates sint nihil magnam', artist='perspiciatis dolorem', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='EXPLICABO', title='sed voluptas porro beatae', artist='architecto a', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='POSSIMUS', title='iure voluptatibus magnam sed', artist='dolor vel', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='POSSIMUS', title='nobis accusamus minus beatae', artist='dicta hic', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='2004', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='POSSIMUS', title='quam odio iure doloribus', artist='nam excepturi', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='2004', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='POSSIMUS', title='at in nisi excepturi', artist='adipisci facere', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='POSSIMUS', title='enim alias repellat tempora', artist='consectetur doloribus', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='2013', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='POSSIMUS', title='reiciendis architecto animi molestias', artist='consequatur magnam', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='2015', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='POSSIMUS', title='deserunt debitis distinctio totam', artist='unde aliquid', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='POSSIMUS', title='cum autem perferendis eveniet', artist='dolores occaecati', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='POSSIMUS', title='nihil dolore sint dicta', artist='quasi voluptates', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='POSSIMUS', title='quos doloremque distinctio laudantium', artist='sed aut', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='POSSIMUS', title='architecto debitis eos deserunt', artist='aspernatur molestiae', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='POSSIMUS', title='perferendis aliquam alias corrupti', artist='quam exercitationem', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='POSSIMUS', title='quas aut vitae ipsum', artist='suscipit repudiandae', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='POSSIMUS', title='deserunt veniam voluptas ab', artist='eligendi cum', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='POSSIMUS', title='at nobis placeat cum', artist='distinctio sequi', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='POSSIMUS', title='doloribus expedita est aut', artist='commodi illum', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='POSSIMUS', title='impedit tempora reiciendis dolorum', artist='minus eum', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='POSSIMUS', title='nemo consequatur fugiat impedit', artist='corporis commodi', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='POSSIMUS', title='cumque distinctio suscipit dolorem', artist='maiores quia', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes=''),
        RivendellCart(cart_number='', cut_number='1', type=CartType.Audio, group_name='POSSIMUS', title='aspernatur ipsa ex atque', artist='cupiditate eveniet', album='', year='', isrc='', isci='', label='', client='', agency='', publisher='', composer='', conductor='', song_id='', user_defined='', description='', outcue='', filename='', length='', start_point='', end_point='', segue_start_point='', segue_end_point='', hook_start_point='', hook_end_point='', talk_start_point='', talk_end_point='', fadeup_point='', fadedown_point='', sched_codes='')
    ]

    mock_rivendell_carts = "mock_rivendell_carts"
    mock_filename = "mock_filename"

    stats_limits = StatisticsLimits()

    ds = DatabaseStatistics(
        rivendell_carts=mock_rivendell_carts, 
        output_filename=mock_filename, 
        stats_limits=stats_limits,
        write_limits=False,
        write_full_statistics=False
    )

    organized_carts = ds._organize_by_rivendell_group(defined_carts)

    assert len(organized_carts.keys()) == 4
    
    LAUDANTIUM_carts = organized_carts["LAUDANTIUM"]
    assert len(LAUDANTIUM_carts) == 10

    ASPERIORES_carts = organized_carts["ASPERIORES"]
    assert len(ASPERIORES_carts) == 10

    EXPLICABO_carts = organized_carts["EXPLICABO"]
    assert len(EXPLICABO_carts) == 20

    POSSIMUS_carts = organized_carts["POSSIMUS"]
    assert len(POSSIMUS_carts) == 20


write_file_params, write_file_ids = \
    generate_true_false_matrix_from_list_of_strings(
        "write_file",
        [
            "write_limits",
            "write_full_statistics"
        ]

    )


@pytest.fixture(scope="function", params=write_file_params, ids=write_file_ids)
def setup_write_file(fs, mocker, request):
    params = request.param

    laudantium_line = pd.Series({
        'Number of Songs': 19127, 
        'Shortest Song Length': '0:00:08', 
        'Longest Song Length': '0:25:34', 
        'Outlier Limits': ('0:01:12', '0:06:04'), 
        'Mean': '0:03:36', 
        'Standard Deviation': '0:00:53', 
        'Lower Bound': '0:02:15', 
        'Number of Songs < Lower Bound': 1243, 
        'Upper Bound': '0:06:15', 
        'Number of Songs > Upper Bound': 501, 
        'Percent of Songs Excluded': 9.1
    })
    laudantium_mock = mocker.Mock(
        to_pandas_series=mocker.Mock(return_value=laudantium_line)
    )

    asperiores_line = pd.Series({
        'Number of Songs': 1778, 
        'Shortest Song Length': '0:00:52', 
        'Longest Song Length': '0:13:11', 
        'Outlier Limits': ('0:00:20', '0:06:32'), 
        'Mean': '0:03:27', 
        'Standard Deviation': '0:01:03', 
        'Lower Bound': '0:02:00', 
        'Number of Songs < Lower Bound': 60, 
        'Upper Bound': '0:06:30', 
        'Number of Songs > Upper Bound': 47, 
        'Percent of Songs Excluded': 6
    })
    asperiores_mock = mocker.Mock(
        to_pandas_series=mocker.Mock(return_value=asperiores_line)
    )

    explicabo_line = pd.Series({
        'Number of Songs': 1223, 
        'Shortest Song Length': '0:00:36', 
        'Longest Song Length': '0:23:10', 
        'Outlier Limits': ('0:00:32', '0:06:36'), 
        'Mean': '0:03:33', 
        'Standard Deviation': '0:01:04', 
        'Lower Bound': '0:02:00', 
        'Number of Songs < Lower Bound': 82, 
        'Upper Bound': '0:06:45', 
        'Number of Songs > Upper Bound': 41, 
        'Percent of Songs Excluded': 10.1
    })
    explicabo_mock = mocker.Mock(
        to_pandas_series=mocker.Mock(return_value=explicabo_line)
    )

    possimus_line = pd.Series({
        'Number of Songs': 4946, 
        'Shortest Song Length': '0:00:12', 
        'Longest Song Length': '0:25:25', 
        'Outlier Limits': ('0:00:26', '0:06:22'), 
        'Mean': '0:03:25', 
        'Standard Deviation': '0:00:58', 
        'Lower Bound': '0:02:00', 
        'Number of Songs < Lower Bound': 157, 
        'Upper Bound': '0:06:15', 
        'Number of Songs > Upper Bound': 140, 
        'Percent of Songs Excluded': 6
    })
    possimus_mock = mocker.Mock(
        to_pandas_series=mocker.Mock(return_value=possimus_line)
    )

    statistics_per_group = {
        "LAUDANTIUM": laudantium_mock,
        "ASPERIORES": asperiores_mock,
        "EXPLICABO": explicabo_mock,
        "POSSIMUS": possimus_mock
    }

    root_dir = Path(r"test")
    root_dir.mkdir()

    mock_rivendell_carts = "mock_rivendell_carts"
    stats_limits = StatisticsLimits()
    ds = DatabaseStatistics(
        rivendell_carts=mock_rivendell_carts, 
        output_filename=None, 
        stats_limits=stats_limits,
        write_limits=params.write_limits,
        write_full_statistics=params.write_full_statistics
    )

    return make_namedtuple(
        "setup_write_file",
        statistics_per_group=statistics_per_group,
        root_dir=root_dir,
        ds=ds,
        params=params
    )


def test__write_csv(setup_write_file):
    ds = setup_write_file.ds

    data_contents = \
        "Group Name,Number of Songs,Shortest Song Length,Longest Song Length,Outlier Limits,Mean,Standard Deviation,Lower Bound,Number of Songs < Lower Bound,Upper Bound,Number of Songs > Upper Bound,Percent of Songs Excluded\n" \
        "ASPERIORES,1778,0:00:52,0:13:11,\"('0:00:20', '0:06:32')\",0:03:27,0:01:03,0:02:00,60,0:06:30,47,6\n" \
        "EXPLICABO,1223,0:00:36,0:23:10,\"('0:00:32', '0:06:36')\",0:03:33,0:01:04,0:02:00,82,0:06:45,41,10.1\n" \
        "LAUDANTIUM,19127,0:00:08,0:25:34,\"('0:01:12', '0:06:04')\",0:03:36,0:00:53,0:02:15,1243,0:06:15,501,9.1\n" \
        "POSSIMUS,4946,0:00:12,0:25:25,\"('0:00:26', '0:06:22')\",0:03:25,0:00:58,0:02:00,157,0:06:15,140,6\n"

    if setup_write_file.params.write_limits:
        expected_file_contents = ",Smallest Standard Deviation,Minimum Population for Outliers,Lower Bound Multiple," \
            "Upper Bound Multiple\nStatistics Limits,0:00:15,4,1.5,3.0\n" + data_contents
    else:
        expected_file_contents = data_contents

    pathname = setup_write_file.root_dir / 'testfile.csv'
    ds.output_filename = pathname
    ds._write_csv(statistics_per_group=setup_write_file.statistics_per_group)

    assert pathname.read_text() == expected_file_contents

    for mock in setup_write_file.statistics_per_group.values():
        mock.to_pandas_series.assert_called_once_with(setup_write_file.params.write_full_statistics)


def test__write_excel(setup_write_file):
    ds = setup_write_file.ds
    pathname = setup_write_file.root_dir / 'testfile.xlsx'
    ds.output_filename = pathname
    ds._write_excel(statistics_per_group=setup_write_file.statistics_per_group)

    with pd.ExcelFile(pathname) as xlsx:
        if setup_write_file.params.write_limits:
            expected_limits = { 
                "Statistics Limits" : {
                        "Smallest Standard Deviation": "0:00:15",
                        "Minimum Population for Outliers": 4,
                        "Lower Bound Multiple": 1.5,
                        "Upper Bound Multiple": 3.0
                }
            }
            df_limits = pd.read_excel(xlsx, sheet_name="Limits", index_col=0).T.to_dict()
            assert df_limits == expected_limits

        expected_data = {
            "ASPERIORES": {
                'Number of Songs': 1778, 
                'Shortest Song Length': '0:00:52', 
                'Longest Song Length': '0:13:11', 
                'Outlier Limits': "('0:00:20', '0:06:32')", 
                'Mean': '0:03:27', 
                'Standard Deviation': '0:01:03', 
                'Lower Bound': '0:02:00', 
                'Number of Songs < Lower Bound': 60, 
                'Upper Bound': '0:06:30', 
                'Number of Songs > Upper Bound': 47, 
                'Percent of Songs Excluded': 6.0
            },
            "EXPLICABO": {
                'Number of Songs': 1223, 
                'Shortest Song Length': '0:00:36', 
                'Longest Song Length': '0:23:10', 
                'Outlier Limits': "('0:00:32', '0:06:36')", 
                'Mean': '0:03:33', 
                'Standard Deviation': '0:01:04', 
                'Lower Bound': '0:02:00', 
                'Number of Songs < Lower Bound': 82, 
                'Upper Bound': '0:06:45', 
                'Number of Songs > Upper Bound': 41, 
                'Percent of Songs Excluded': 10.1
            },
            "LAUDANTIUM": {
                'Number of Songs': 19127, 
                'Shortest Song Length': '0:00:08', 
                'Longest Song Length': '0:25:34', 
                'Outlier Limits': "('0:01:12', '0:06:04')", 
                'Mean': '0:03:36', 
                'Standard Deviation': '0:00:53', 
                'Lower Bound': '0:02:15', 
                'Number of Songs < Lower Bound': 1243, 
                'Upper Bound': '0:06:15', 
                'Number of Songs > Upper Bound': 501, 
                'Percent of Songs Excluded': 9.1
            },
            "POSSIMUS": {
                'Number of Songs': 4946, 
                'Shortest Song Length': '0:00:12', 
                'Longest Song Length': '0:25:25', 
                'Outlier Limits': "('0:00:26', '0:06:22')", 
                'Mean': '0:03:25', 
                'Standard Deviation': '0:00:58', 
                'Lower Bound': '0:02:00', 
                'Number of Songs < Lower Bound': 157, 
                'Upper Bound': '0:06:15', 
                'Number of Songs > Upper Bound': 140, 
                'Percent of Songs Excluded': 6.0
            }
        }

        df_data = pd.read_excel(xlsx, sheet_name="Data", index_col=0).T.to_dict()

        assert df_data == expected_data

    for mock in setup_write_file.statistics_per_group.values():
        mock.to_pandas_series.assert_called_once_with(setup_write_file.params.write_full_statistics)


run_script_params, run_script_ids = \
    generate_true_false_matrix_from_list_of_strings(
        "run_script",
        [
            "file_already_exists",
            "excel_file"
        ]
)

@pytest.mark.parametrize("params", run_script_params, ids=run_script_ids)
def test_run_script(fs, params, mocker):
    rivendell_carts = "mock_rivendell_carts"

    if params.excel_file:
        output_filename = Path("/rivendell/output_filename.xlsx")
    else:
        output_filename = Path("/rivendell/output_filename.csv")
    
    if params.file_already_exists:
        fs.create_file(output_filename)
        assert output_filename.exists()
    else:
        assert not output_filename.exists()

    mock_organized_by_rivendell_group = "mock_organized_by_rivendell_group"

    mock_organize_by_rivendell_groups_function = mocker.Mock(
        return_value=mock_organized_by_rivendell_group
    )

    mock_statistics_per_group = "mock_statistics_per_group"
    mock_calculate_statistics_per_group_function = mocker.Mock(
        return_value=mock_statistics_per_group
    )

    mock_write_excel = mocker.Mock()
    mock_write_csv = mocker.Mock()

    ds = mocker.Mock(
        rivendell_carts=rivendell_carts,
        output_filename=output_filename,
        _organize_by_rivendell_group=mock_organize_by_rivendell_groups_function,
        _calculate_statistics_per_group=mock_calculate_statistics_per_group_function,
        _write_excel=mock_write_excel,
        _write_csv=mock_write_csv
    )

    result = DatabaseStatistics.run_script(ds)

    assert result is None

    mock_organize_by_rivendell_groups_function.assert_called_once_with(
        unorganized_carts=rivendell_carts
    )

    mock_calculate_statistics_per_group_function.assert_called_once_with(
        organized_carts=mock_organized_by_rivendell_group
    )

    if params.file_already_exists:
        new_filename = (output_filename.parent / 
            (output_filename.stem + "_old" + output_filename.suffix))
        assert new_filename.exists()
    assert not output_filename.exists()

    if params.excel_file:
        mock_write_excel.assert_called_once_with(
            statistics_per_group=mock_statistics_per_group
        )
    else:
        mock_write_csv.assert_called_once_with(
            statistics_per_group=mock_statistics_per_group
        )



"""
@Author = 'Michael Stanley'

============ Change Log ============
2023-Jan-20 = Change license from GPLv2 to GPLv3.

2023-Jan-19 = Created.

============ License ============
Copyright (C) 2023, 2025 Michael Stanley

This file is part of wmul_rivendell.

wmul_rivendell is free software: you can redistribute it and/or modify it 
under the terms of the GNU General Public License as published by the Free 
Software Foundation, either version 3 of the License, or (at your option) any 
later version.

wmul_rivendell is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with 
wmul_rivendell. If not, see <https://www.gnu.org/licenses/>. 
"""
from click.testing import CliRunner
import pytest

from wmul_rivendell import cli
from wmul_test_utils import generate_true_false_matrix_from_list_of_strings

filter_cart_report_params, filter_cart_report_ids = \
    generate_true_false_matrix_from_list_of_strings(
        "filter_cart_report",
        [
            "include_macros",
            "include_all_cuts",
            "use_trailing_comma",
            "exclude_groups"
        ]

    )

@pytest.mark.parametrize("params", filter_cart_report_params, ids=filter_cart_report_ids)
def test_filter_cart_report(fs, params, mocker, caplog):
    mock_rivendell_cart_filename = "/test/mock_rivendell_cart_filename.txt"
    fs.create_file(mock_rivendell_cart_filename)
    mock_output_filename = "/test/mock_output_filename"
    mock_desired_fields_filename = "/test/desired.txt"

    desired_fields_file_contents = "Cart_number\nGroup_Name\nTitle\nArtist\nAlbum\nYear\nLength\nUser_Defined\nSched_Codes\n"
    expected_desired_fields = [
        "Cart_number", 
        "Group_Name",
        "Title",
        "Artist",
        "Album",
        "Year",
        "Length",
        "User_Defined",
        "Sched_Codes"
    ]

    fs.create_file(mock_desired_fields_filename, 
                   contents=desired_fields_file_contents)

    mock_rivendell_carts = "mock_rivendell_carts"
    mock_load_carts = mocker.Mock(return_value=mock_rivendell_carts)

    mock_load_cart_data_dump_object = mocker.Mock(load_carts=mock_load_carts)
    mock_load_cart_data_dump_constructor = mocker.patch(
        "wmul_rivendell.cli.LoadCartDataDump",
        return_value=mock_load_cart_data_dump_object,
        autospec=True
    )

    mock_filter_cart_report_object = mocker.Mock()

    mock_filter_cart_report_constructor = mocker.patch(
        "wmul_rivendell.cli.FilterCartReportForMusicScheduler",
        return_value=mock_filter_cart_report_object,
        autospec=True
    )

    cli_args = [
            mock_rivendell_cart_filename,
            mock_output_filename,
            mock_desired_fields_filename
        ]
    
    if params.include_macros:
        cli_args.append("--include_macros")
    
    if params.include_all_cuts:
        cli_args.append("--include_all_cuts")

    if params.use_trailing_comma:
        cli_args.append("--use_trailing_comma")

    if params.exclude_groups:
        exclude_groups_file_contents = "LAUDANTIUM\nASPERIORES\n"
        mock_exclude_groups_filename = "/test/mock_exclude_groups_filename.txt"
        fs.create_file(mock_exclude_groups_filename, contents=exclude_groups_file_contents)
        cli_args.append("--excluded_groups_file_name")
        cli_args.append(mock_exclude_groups_filename)
        expected_exclude_groups = ["LAUDANTIUM", "ASPERIORES"]
    else:
        expected_exclude_groups = []

    runner = CliRunner()
    result = runner.invoke(
        cli.filter_cart_report,
        cli_args
    )

    assert result.exit_code == 0

    mock_load_cart_data_dump_constructor.assert_called_once_with(
        rivendell_cart_data_filename=mock_rivendell_cart_filename,
        include_macros=params.include_macros,
        include_all_cuts=params.include_all_cuts,
        excluded_group_list=expected_exclude_groups,
    )

    mock_load_carts.assert_called_once_with()

    mock_filter_cart_report_constructor.assert_called_once_with(
        rivendell_carts=mock_rivendell_carts,
        output_filename=mock_output_filename,
        desired_field_list=expected_desired_fields,
        use_trailing_comma=params.use_trailing_comma
    )

    mock_filter_cart_report_object.run_script.assert_called_once_with()
